# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018, 2019 . All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

import struct
import numpy as np
from scipy import sparse, stats

def read_from_snap_format(filename):
    """
    Non-distributed data loading from snap format

    Parameters
    ----------
    filename : str
        The file where the data resides.

    Returns
    ------- 
    X, y : numpy array or sparse matrix, numpy array
        Returns two datasets. 
        X : the data used for training or inference
        y : the labels of the samples in X. 
    """
    with open(filename, "rb") as f:

        data_type = struct.unpack('I',f.read(4))[0]

        if(data_type == 0):
            # dense data
            print("Reading dense data from snap format")
            transpose = struct.unpack('?',f.read(1))[0]
            num_ex    = struct.unpack('I',f.read(4))[0]
            num_ft    = struct.unpack('I',f.read(4))[0]
            y         = np.frombuffer(f.read(4*num_ex),dtype=np.float32)
            X         = np.frombuffer(f.read(4*num_ex*num_ft),dtype=np.float32)

            if transpose:
                X.shape = (num_ft,num_ex)
            else:
                X.shape = (num_ex,num_ft)

            return X, y

        elif(data_type == 1):
            # sparse data
            print("Reading sparse data from snap format")
            transpose = struct.unpack('?',f.read(1))[0]
            num_ex    = struct.unpack('I',f.read(4))[0]
            num_ft    = struct.unpack('I',f.read(4))[0]

            if transpose:
                count = np.frombuffer(f.read(4*num_ft),dtype=np.uint32)
            else:
                count = np.frombuffer(f.read(4*num_ex),dtype=np.uint32)

            num_nz = np.sum(count)
            indptr = np.cumsum(np.insert(count,0,0))
            num_bytes = 4*num_nz
            y      = np.frombuffer(f.read(4*num_ex), dtype=np.float32)
            ind    = np.frombuffer(f.read(num_bytes.astype(np.uint64)), dtype=np.uint32)
            val    = np.frombuffer(f.read(num_bytes.astype(np.uint64)), dtype=np.float32)

            if transpose:
                X = sparse.csc.csc_matrix((val,ind,indptr),shape=(num_ex,num_ft))
            else:
                X = sparse.csr.csr_matrix((val,ind,indptr),shape=(num_ex,num_ft))

            return X, y

        elif(data_type == 2):
            # l2-sparse data
            raise Exception("Sparse data with implicit values not supported by numpy/scipy")
        else:
            # not recognized
            raise Exception("Unrecognized data type")
	
		


def dump_to_snap_format(X, y, filename, transpose=False, implicit_vals=False):
    """
    Non-distributed data writing to snap format

    Parameters
    ----------
    X : numpy array or sparse matrix
        The data used for training or inference.             

    y : numpy array
        The labels of the samples in X.

    filename : str
        The file where X and y will be stored in snap format.

    transpose : bool , default : False
        If transpose is True, X will be stored in transposed format.
    """

    num_ex = X.shape[0]
    num_ft = X.shape[1]
    assert y.shape[0] == num_ex

    assert type(y) is np.ndarray
    y = y.astype(np.float32)

    if(type(X) is np.ndarray):
        print("Writing Dense dataset to snap format")

        # deep copy 
        if transpose:
            X = X.T.copy()

        X = X.astype(np.float32)

        with open(filename, "wb") as f:
            f.write(struct.pack('I', 0))
            f.write(struct.pack('?', transpose))
            f.write(struct.pack('I', num_ex))
            f.write(struct.pack('I', num_ft))
            f.write(y.data)
            f.write(X.data)

    elif(type(X) is sparse.csr.csr_matrix):

        print("Writing Sparse dataset to snap format")
        if transpose:
            X = X.tocsc()

        ind   = X.indices.astype(np.uint32) 
        val   = X.data.astype(np.float32)
        count = np.diff(X.indptr).astype(np.uint32)

        with open(filename, "wb") as f:
            if implicit_vals:
                f.write(struct.pack('I', 2))
            else:
                f.write(struct.pack('I', 1))
            f.write(struct.pack('?', transpose))
            f.write(struct.pack('I', num_ex))
            f.write(struct.pack('I', num_ft))
            f.write(count.data)
            f.write(y.data)
            f.write(ind.data)

            if not implicit_vals:
                f.write(val.data)
    else:
	    print("Unrecognized type; cannot write to snap format")

