# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# ******************************************************************

from pyspark import keyword_only
from pyspark.ml.param.shared import *
from pyspark.ml.util import *
from pyspark.ml.wrapper import JavaEstimator, JavaModel, JavaParams
from pyspark.ml.wrapper import JavaWrapper
from pyspark.ml.common import inherit_doc, _py2java, _picklable_classes
from pyspark.ml import common
from py4j.java_gateway import JavaObject
from py4j.java_collections import JavaArray, JavaList
from pyspark import RDD, SparkContext
from pyspark.sql import DataFrame, SQLContext
from pyspark.serializers import PickleSerializer, AutoBatchedSerializer

__all__ = ['LinearSVC', 'LinearSVCModel', 'LinearSVCSummary',
           'LogisticRegression', 'LogisticRegressionModel',
           'LogisticRegressionSummary', 'LogisticRegressionTrainingSummary',
           'BinaryLogisticRegressionSummary', 'BinaryLogisticRegressionTrainingSummary',
           'LinearRegression', 'LinearRegressionModel',
           'LinearRegressionSummary', 'LinearRegressionTrainingSummary']

def _java2py_internal(sc, r, encoding="bytes"):
    if isinstance(r, JavaObject):
        clsName = r.getClass().getSimpleName()
        # convert RDD into JavaRDD
        if clsName != 'JavaRDD' and clsName.endswith("RDD"):
            r = r.toJavaRDD()
            clsName = 'JavaRDD'

        if clsName == 'JavaRDD':
            jrdd = sc._jvm.org.apache.spark.ml.python.MLSerDe.javaToPython(r)
            return RDD(jrdd, sc)

        if clsName == 'Dataset':
            return DataFrame(r, SQLContext.getOrCreate(sc))

        if clsName in _picklable_classes:
            r = sc._jvm.org.apache.spark.ml.python.MLSerDe.dumps(r)
        elif isinstance(r, (JavaArray, JavaList)):
            try:
                r = sc._jvm.org.apache.spark.ml.python.MLSerDe.dumps(r)
            except Py4JJavaError:
                pass  # not pickable

    if isinstance(r, (bytearray, bytes)):
        r = PickleSerializer().loads(bytes(r), encoding=encoding)
    return r

def _call_java_internal(self, name, *args):
    m = getattr(self._java_obj, name)
    sc = SparkContext._active_spark_context
    java_args = [_py2java(sc, arg) for arg in args]
    if name == "objectiveHistory":
        histories = m(*java_args)
        training_history_ = []
        for history in histories:
            if isinstance(history, (JavaArray, JavaList)):
                history = sc._jvm.org.apache.spark.ml.python.MLSerDe.dumps(history)
            if isinstance(history, (bytearray, bytes)):
                history = PickleSerializer().loads(bytes(history), encoding="bytes")
            training_history_.append(history)
        return training_history_   
    else:    
        return _java2py_internal(sc, m(*java_args))

JavaWrapper._call_java = _call_java_internal

class _tmpClass1(JavaMLReader):
    @classmethod
    def _java_loader_class_override(cls, clazz):
        """
        Returns the full class name of the Java ML instance. The default
        implementation replaces "pyspark" by "org.apache.spark" in
        the Python full class name.
        """
        if "snap_ml_spark" in clazz.__module__:
            java_package = clazz.__module__.replace("snap_ml_spark.estimator", "com.ibm.snap.ml")
        else:
            java_package = clazz.__module__.replace("pyspark", "org.apache.spark")
        if clazz.__name__ in ("Pipeline", "PipelineModel"):
            # Remove the last package name "pipeline" for Pipeline and PipelineModel.
            java_package = ".".join(java_package.split(".")[0:-1])
        if "snap_ml_spark" in clazz.__module__:
            return java_package + ".Snap" + clazz.__name__
        else:
            return java_package + "." + clazz.__name__

JavaMLReader._java_loader_class = _tmpClass1._java_loader_class_override

class _tmpClass2(JavaParams):
    @classmethod
    def _from_java_override(cls, java_stage):
        """
        Given a Java object, create and return a Python wrapper of it.
        Used for ML persistence.

        Meta-algorithms such as Pipeline should override this method as a classmethod.
        """
        def __get_class(clazz):
            """
            Loads Python class from its name.
            """
            parts = clazz.split('.')
            module = ".".join(parts[:-1])
            m = __import__(module)
            for comp in parts[1:]:
                m = getattr(m, comp)
            return m
        if "com.ibm.snap.ml" in java_stage.getClass().getName():
            stage_name = java_stage.getClass().getName().replace("com.ibm.snap.ml.Snap", "snap_ml_spark.estimator.")
        else:
            stage_name = java_stage.getClass().getName().replace("org.apache.spark", "pyspark")
        # Generate a default new instance from the stage_name class.
        py_type = __get_class(stage_name)
        if issubclass(py_type, JavaParams):
            # Load information from java_stage to the instance.
            py_stage = py_type()
            py_stage._java_obj = java_stage

            # SPARK-10931: Temporary fix so that persisted models would own params from Estimator
            if issubclass(py_type, JavaModel):
                py_stage._create_params_from_java()

            py_stage._resetUid(java_stage.uid())
            py_stage._transfer_params_from_java()
        elif hasattr(py_type, "_from_java"):
            py_stage = py_type._from_java(java_stage)
        else:
            raise NotImplementedError("This Java stage cannot be loaded into Python currently: %r"
                                      % stage_name)
        return py_stage


JavaParams._from_java = _tmpClass2._from_java_override

@inherit_doc
class LinearRegression(JavaEstimator, HasFeaturesCol, HasLabelCol, HasPredictionCol, HasMaxIter,
                       HasRegParam, HasTol, HasElasticNetParam, HasWeightCol, JavaMLWritable, JavaMLReadable):
    """
    Linear regression.

    The learning objective is to minimize the specified loss function, with regularization.


    >>> from pyspark.ml.linalg import Vectors
    >>> df = spark.createDataFrame([
    ...     (1.0, 2.0, Vectors.dense(1.0)),
    ...     (0.0, 2.0, Vectors.sparse(1, [], []))], ["label", "weight", "features"])
    >>> lr = LinearRegression(maxIter=5, regParam=0.0, weightCol="weight")
    >>> model = lr.fit(df)
    >>> test0 = spark.createDataFrame([(Vectors.dense(-1.0),)], ["features"])
    >>> abs(model.transform(test0).head().prediction - (-1.0)) < 0.001
    True
    >>> abs(model.coefficients[0] - 1.0) < 0.001
    True
    >>> abs(model.intercept - 0.0) < 0.001
    True
    >>> test1 = spark.createDataFrame([(Vectors.sparse(1, [0], [1.0]),)], ["features"])
    >>> abs(model.transform(test1).head().prediction - 1.0) < 0.001
    True
    >>> lr.setParams("vector")
    Traceback (most recent call last):
        ...
    TypeError: Method setParams forces keyword arguments.
    >>> lr_path = temp_path + "/lr"
    >>> lr.save(lr_path)
    >>> lr2 = LinearRegression.load(lr_path)
    >>> lr2.getMaxIter()
    5
    >>> model_path = temp_path + "/lr_model"
    >>> model.save(model_path)
    >>> model2 = LinearRegressionModel.load(model_path)
    >>> model.coefficients[0] == model2.coefficients[0]
    True
    >>> model.intercept == model2.intercept
    True
    >>> model.numFeatures
    1

    """

    verbose = Param(Params._dummy(), "verbose", "Flag for indicating if the training loss will be printed at each epoch.",
                    typeConverter=TypeConverters.toBoolean)

    def setVerbose(self, value):
        """
        Sets the value of :py:attr:`verbose`.
        """
        return self._set(verbose=value)

    def getVerbose(self):
        """
        Gets the value of verbose or its default value.
        """
        return self.getOrDefault(self.verbose)

    gpuMemLimit = Param(Params._dummy(), "gpuMemLimit", "Limit of the GPU memory. If set to the default value 0, the maximum possible memory is used.",
                    typeConverter=TypeConverters.toInt)

    def setGpuMemLimit(self, value):
        """
        Sets the value of :py:attr:`gpuMemLimit`.
        """
        return self._set(gpuMemLimit=value)

    def getGpuMemLimit(self):
        """
        Gets the value of gpuMemLimit or its default value.
        """
        return self.getOrDefault(self.gpuMemLimit)

    nthreads = Param(Params._dummy(), "nthreads", "Number of threads to be used. default : -1 meaning that n_threads=256 if GPU is enabled, else 1",
                        typeConverter=TypeConverters.toInt)

    def setNthreads(self, value):
        """
        Sets the value of :py:attr:`nthreads`.
        """
        return self._set(nthreads=value)

    def getNthreads(self):
        """
        Gets the value of nthreads or its default value.
        """
        return self.getOrDefault(self.nthreads)

    balanced = Param(Params._dummy(), "balanced", "If set to False, all classes will have weight 1.",
                     typeConverter=TypeConverters.toBoolean)

    def setBalanced(self, value):
        """
        Sets the value of :py:attr:`balanced`.
        """
        return self._set(balanced=value)

    def getBalanced(self):
        """
        Gets the value of balanced or its default value.
        """
        return self.getOrDefault(self.balanced)

    dual = Param(Params._dummy(), "dual", "Dual or Primal formulation. Recommendation: if n_samples > n_features use dual=True.",
                     typeConverter=TypeConverters.toBoolean)

    def setDual(self, value):
        """
        Sets the value of :py:attr:`dual`.
        """
        return self._set(dual=value)

    def getDual(self):
        """
        Gets the value of dual or its default value.
        """
        return self.getOrDefault(self.dual)

    useGpu = Param(Params._dummy(), "useGpu", "Flag for indicating the hardware platform used for training. If True, the training is performed using the GPU. If False, the training is performed using the CPU.",
                 typeConverter=TypeConverters.toBoolean)

    def setUseGpu(self, value):
        """
        Sets the value of :py:attr:`useGpu`.
        """
        return self._set(useGpu=value)

    def getUseGpu(self):
        """
        Gets the value of useGpu or its default value.
        """
        return self.getOrDefault(self.useGpu)

    trainingHistory = Param(Params._dummy(), "trainingHistory", "Flag for indicating the amount of information to collect and returned by the fit function. default: 0, no information is returned, 1, to obtain summary statistics at the end of training, 2, to obtain a complete set of statistics for the entire training procedure",
                 typeConverter=TypeConverters.toInt)

    def setTrainingHistory(self, value):
        """
        Sets the value of :py:attr:`trainingHistory`.
        """
        return self._set(trainingHistory=value)

    def getTrainingHistory(self):
        """
        Gets the value of trainingHistory or its default value.
        """
        return self.getOrDefault(self.trainingHistory)

    @keyword_only
    def __init__(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0,
                 maxIter=1000, regParam=0.0, elasticNetParam=0.0, tol=1e-3, solver="auto", weightCol=None,
                 useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):
        """
        __init__(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0, \
                 maxIter=1000, regParam=0.0, elasticNetParam=0.0, tol=1e-3, solver="auto", weightCol=None, \
                 useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False)
        """
        super(LinearRegression, self).__init__()
        self._java_obj = self._new_java_obj(
            "com.ibm.snap.ml.SnapLinearRegression", self.uid)
        self._setDefault(trainingHistory=0, maxIter=1000, regParam=1.0, tol=1e-3, useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0,
                  maxIter=1000, regParam=0.0, elasticNetParam=0.0, tol=1e-3, solver="auto", weightCol=None,
                  useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):
        """
        setParams(featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0, \
                 maxIter=1000, regParam=0.0, elasticNetParam=0.0, tol=1e-6, solver="auto", weightCol=None, \
                 useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False)
        Sets params for linear regression.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def _create_model(self, java_model):
        return LinearRegressionModel(java_model)




class LinearRegressionModel(JavaModel, JavaPredictionModel, JavaMLWritable, JavaMLReadable):
    """
    Model fitted by :class:`LinearRegression`.

    """

    @property
    def coefficients(self):
        """
        Model coefficients.
        """
        return self._call_java("coefficients")

    @property
    def intercept(self):
        """
        Model intercept.
        """
        return self._call_java("intercept")

    @property
    def scale(self):
        """
        The value will be 1.0.
        """
        return self._call_java("scale")

    @property
    def summary(self):
        """
        Gets summary (e.g. residuals, mse, r-squared ) of model on
        training set. An exception is thrown if
        `trainingSummary is None`.
        """
        if self.hasSummary:
            java_lrt_summary = self._call_java("summary")
            return LinearRegressionTrainingSummary(java_lrt_summary)
        else:
            raise RuntimeError("No training summary available for this %s" %
                               self.__class__.__name__)

    @property
    def hasSummary(self):
        """
        Indicates whether a training summary exists for this model
        instance.
        """
        return self._call_java("hasSummary")

    def evaluate(self, dataset):
        """
        Evaluates the model on a test dataset.

        :param dataset:
          Test dataset to evaluate model on, where dataset is an
          instance of :py:class:`pyspark.sql.DataFrame`
        """
        if not isinstance(dataset, DataFrame):
            raise ValueError("dataset must be a DataFrame but got %s." % type(dataset))
        java_lr_summary = self._call_java("evaluate", dataset)
        return LinearRegressionSummary(java_lr_summary)


class LinearRegressionSummary(JavaWrapper):
    """

    Linear regression results evaluated on a dataset.

    """

    @property
    def predictions(self):
        """
        Dataframe outputted by the model's `transform` method.
        """
        return self._call_java("predictions")

    @property
    def predictionCol(self):
        """
        Field in "predictions" which gives the predicted value of
        the label at each instance.
        """
        return self._call_java("predictionCol")

    @property
    def labelCol(self):
        """
        Field in "predictions" which gives the true label of each
        instance.
        """
        return self._call_java("labelCol")

    @property
    def featuresCol(self):
        """
        Field in "predictions" which gives the features of each instance
        as a vector.
        """
        return self._call_java("featuresCol")

    @property
    def explainedVariance(self):
        """
        Returns the explained variance regression score.
        explainedVariance = 1 - variance(y - \hat{y}) / variance(y)

        .. seealso:: `Wikipedia explain variation \
        <http://en.wikipedia.org/wiki/Explained_variation>`_

        """
        return self._call_java("explainedVariance")

    @property
    def meanAbsoluteError(self):
        """
        Returns the mean absolute error, which is a risk function
        corresponding to the expected value of the absolute error
        loss or l1-norm loss.

        """
        return self._call_java("meanAbsoluteError")

    @property
    def meanSquaredError(self):
        """
        Returns the mean squared error, which is a risk function
        corresponding to the expected value of the squared error
        loss or quadratic loss.

        """
        return self._call_java("meanSquaredError")

    @property
    def rootMeanSquaredError(self):
        """
        Returns the root mean squared error, which is defined as the
        square root of the mean squared error.

        """
        return self._call_java("rootMeanSquaredError")

    @property
    def r2(self):
        """
        Returns R^2^, the coefficient of determination.

        .. seealso:: `Wikipedia coefficient of determination \
        <http://en.wikipedia.org/wiki/Coefficient_of_determination>`

        """
        return self._call_java("r2")

    @property
    def residuals(self):
        """
        Residuals (label - predicted value)
        """
        return self._call_java("residuals")

    @property
    def numInstances(self):
        """
        Number of instances in DataFrame predictions
        """
        return self._call_java("numInstances")

    @property
    def degreesOfFreedom(self):
        """
        Degrees of freedom.
        """
        return self._call_java("degreesOfFreedom")

    @property
    def devianceResiduals(self):
        """
        The weighted residuals, the usual residuals rescaled by the
        square root of the instance weights.
        """
        return self._call_java("devianceResiduals")

    @property
    def coefficientStandardErrors(self):
        """
        Standard error of estimated coefficients and intercept.
        This value is only available when using the "normal" solver.

        If :py:attr:`LinearRegression.fitIntercept` is set to True,
        then the last element returned corresponds to the intercept.

        .. seealso:: :py:attr:`LinearRegression.solver`
        """
        return self._call_java("coefficientStandardErrors")

    @property
    def tValues(self):
        """
        T-statistic of estimated coefficients and intercept.
        This value is only available when using the "normal" solver.

        If :py:attr:`LinearRegression.fitIntercept` is set to True,
        then the last element returned corresponds to the intercept.

        .. seealso:: :py:attr:`LinearRegression.solver`
        """
        return self._call_java("tValues")

    @property
    def pValues(self):
        """
        Two-sided p-value of estimated coefficients and intercept.
        This value is only available when using the "normal" solver.

        If :py:attr:`LinearRegression.fitIntercept` is set to True,
        then the last element returned corresponds to the intercept.

        .. seealso:: :py:attr:`LinearRegression.solver`
        """
        return self._call_java("pValues")


@inherit_doc
class LinearRegressionTrainingSummary(LinearRegressionSummary):
    """

    Linear regression training results. Currently, the training summary ignores the
    training weights except for the objective trace.

    """

    @property
    def objectiveHistory(self):
        """
        Objective function (scaled loss + regularization) at each
        iteration.

        .. seealso:: :py:attr:`LinearRegression.solver`
        """
        return self._call_java("objectiveHistory")

    @property
    def totalIterations(self):
        """
        Number of training iterations until termination.

        .. seealso:: :py:attr:`LinearRegression.solver`
        """
        return self._call_java("totalIterations")

class LogisticRegression(JavaEstimator, HasFeaturesCol, HasLabelCol, HasPredictionCol, HasMaxIter,
                         HasRegParam, HasTol, HasProbabilityCol, HasRawPredictionCol,
                         HasElasticNetParam, JavaMLWritable, JavaMLReadable):
    """
    Logistic regression.
    This class supports multinomial logistic (softmax) and binomial logistic regression.

    >>> from pyspark.sql import Row
    >>> from pyspark.ml.linalg import Vectors
    >>> bdf = sc.parallelize([
    ...     Row(label=1.0, weight=1.0, features=Vectors.dense(0.0, 5.0)),
    ...     Row(label=0.0, weight=2.0, features=Vectors.dense(1.0, 2.0)),
    ...     Row(label=1.0, weight=3.0, features=Vectors.dense(2.0, 1.0)),
    ...     Row(label=0.0, weight=4.0, features=Vectors.dense(3.0, 3.0))]).toDF()
    >>> blor = LogisticRegression(regParam=0.01, weightCol="weight")
    >>> blorModel = blor.fit(bdf)
    >>> blorModel.coefficients
    DenseVector([-1.080..., -0.646...])
    >>> blorModel.intercept
    3.112...
    >>> data_path = "data/mllib/sample_multiclass_classification_data.txt"
    >>> mdf = spark.read.format("libsvm").load(data_path)
    >>> mlor = LogisticRegression(regParam=0.1, elasticNetParam=1.0, family="multinomial")
    >>> mlorModel = mlor.fit(mdf)
    >>> mlorModel.coefficientMatrix
    SparseMatrix(3, 4, [0, 1, 2, 3], [3, 2, 1], [1.87..., -2.75..., -0.50...], 1)
    >>> mlorModel.interceptVector
    DenseVector([0.04..., -0.42..., 0.37...])
    >>> test0 = sc.parallelize([Row(features=Vectors.dense(-1.0, 1.0))]).toDF()
    >>> result = blorModel.transform(test0).head()
    >>> result.prediction
    1.0
    >>> result.probability
    DenseVector([0.02..., 0.97...])
    >>> result.rawPrediction
    DenseVector([-3.54..., 3.54...])
    >>> test1 = sc.parallelize([Row(features=Vectors.sparse(2, [0], [1.0]))]).toDF()
    >>> blorModel.transform(test1).head().prediction
    1.0
    >>> blor.setParams("vector")
    Traceback (most recent call last):
        ...
    TypeError: Method setParams forces keyword arguments.
    >>> lr_path = temp_path + "/lr"
    >>> blor.save(lr_path)
    >>> lr2 = LogisticRegression.load(lr_path)
    >>> lr2.getRegParam()
    0.01
    >>> model_path = temp_path + "/lr_model"
    >>> blorModel.save(model_path)
    >>> model2 = LogisticRegressionModel.load(model_path)
    >>> blorModel.coefficients[0] == model2.coefficients[0]
    True
    >>> blorModel.intercept == model2.intercept
    True

    """

    family = Param(Params._dummy(), "family",
                   "The name of family which is a description of the label distribution to " +
                   "be used in the model. Supported options: auto, binomial, multinomial",
                   typeConverter=TypeConverters.toString)

    def setFamily(self, value):
        """
        Sets the value of :py:attr:`family`.
        """
        return self._set(family=value)

    def getFamily(self):
        """
        Gets the value of :py:attr:`family` or its default value.
        """
        return self.getOrDefault(self.family)

    verbose = Param(Params._dummy(), "verbose", "Flag for indicating if the training loss will be printed at each epoch.",
                    typeConverter=TypeConverters.toBoolean)

    def setVerbose(self, value):
        """
        Sets the value of :py:attr:`verbose`.
        """
        return self._set(verbose=value)

    def getVerbose(self):
        """
        Gets the value of verbose or its default value.
        """
        return self.getOrDefault(self.verbose)

    gpuMemLimit = Param(Params._dummy(), "gpuMemLimit", "Limit of the GPU memory. If set to the default value 0, the maximum possible memory is used.",
                        typeConverter=TypeConverters.toInt)

    def setGpuMemLimit(self, value):
        """
        Sets the value of :py:attr:`gpuMemLimit`.
        """
        return self._set(gpuMemLimit=value)

    def getGpuMemLimit(self):
        """
        Gets the value of gpuMemLimit or its default value.
        """
        return self.getOrDefault(self.gpuMemLimit)

    nthreads = Param(Params._dummy(), "nthreads", "Number of threads to be used. default : -1 meaning that n_threads=256 if GPU is enabled, else 1",
                     typeConverter=TypeConverters.toInt)

    def setNthreads(self, value):
        """
        Sets the value of :py:attr:`nthreads`.
        """
        return self._set(nthreads=value)

    def getNthreads(self):
        """
        Gets the value of nthreads or its default value.
        """
        return self.getOrDefault(self.nthreads)

    balanced = Param(Params._dummy(), "balanced", "If set to False, all classes will have weight 1.",
                     typeConverter=TypeConverters.toBoolean)

    def setBalanced(self, value):
        """
        Sets the value of :py:attr:`balanced`.
        """
        return self._set(balanced=value)

    def getBalanced(self):
        """
        Gets the value of balanced or its default value.
        """
        return self.getOrDefault(self.balanced)

    dual = Param(Params._dummy(), "dual", "Dual or Primal formulation. Recommendation: if n_samples > n_features use dual=True.",
                 typeConverter=TypeConverters.toBoolean)

    def setDual(self, value):
        """
        Sets the value of :py:attr:`dual`.
        """
        return self._set(dual=value)

    def getDual(self):
        """
        Gets the value of dual or its default value.
        """
        return self.getOrDefault(self.dual)

    useGpu = Param(Params._dummy(), "useGpu", "Flag for indicating the hardware platform used for training. If True, the training is performed using the GPU. If False, the training is performed using the CPU.",
                   typeConverter=TypeConverters.toBoolean)

    def setUseGpu(self, value):
        """
        Sets the value of :py:attr:`useGpu`.
        """
        return self._set(useGpu=value)

    def getUseGpu(self):
        """
        Gets the value of useGpu or its default value.
        """
        return self.getOrDefault(self.useGpu)

    trainingHistory = Param(Params._dummy(), "trainingHistory", "Flag for indicating the amount of information to collect and returned by the fit function. default: 0, no information is returned, 1, to obtain summary statistics at the end of training, 2, to obtain a complete set of statistics for the entire training procedure",
                            typeConverter=TypeConverters.toInt)

    def setTrainingHistory(self, value):
        """
        Sets the value of :py:attr:`trainingHistory`.
        """
        return self._set(trainingHistory=value)

    def getTrainingHistory(self):
        """
        Gets the value of trainingHistory or its default value.
        """
        return self.getOrDefault(self.trainingHistory)

    @keyword_only
    def __init__(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0,
                 maxIter=1000, regParam=1.0, elasticNetParam=0.0, probabilityCol="probability",
                 rawPredictionCol="rawPrediction",family="auto", tol=1e-3,
                 useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):

        """
        __init__(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0,\
                 maxIter=1000, regParam=1.0, elasticNetParam=0.0, tol=1e-3,\
                 probabilityCol="probability", rawPredictionCol="rawPrediction", family="auto", \
                 useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):
        If the threshold and thresholds Params are both set, they must be equivalent.
        """
        super(LogisticRegression, self).__init__()
        self._java_obj = self._new_java_obj(
            "com.ibm.snap.ml.SnapLogisticRegression", self.uid)
        self._setDefault(trainingHistory=0, maxIter=1000, regParam=1.0, tol=1E-3, family="auto", useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0,
                  maxIter=1000, regParam=1.0, elasticNetParam=0.0, tol=1e-3,
                  probabilityCol="probability", rawPredictionCol="rawPrediction", family="auto",
                  useGpu=False, dual=True, balanced=False, nthreads=1, gpuMemLimit=0, verbose=False):
        """
        setParams(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0 \
                  maxIter=1000, regParam=1.0, elasticNetParam=0.0, tol=1e-3,\
                  probabilityCol="probability", rawPredictionCol="rawPrediction", family="auto", \
                  useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):
        Sets params for logistic regression.
        If the threshold and thresholds Params are both set, they must be equivalent.
        """
        kwargs = self._input_kwargs
        self._set(**kwargs)
        return self

    def _create_model(self, java_model):
        return LogisticRegressionModel(java_model)


class JavaClassificationModel(JavaPredictionModel):
    """
    (Private) Java Model produced by a ``Classifier``.
    Classes are indexed {0, 1, ..., numClasses - 1}.
    To be mixed in with class:`pyspark.ml.JavaModel`
    """

    @property
    def numClasses(self):
        """
        Number of classes (values which the label can take).
        """
        return self._call_java("numClasses")

class LogisticRegressionModel(JavaModel, JavaClassificationModel, JavaMLWritable, JavaMLReadable):
    """
    Model fitted by LogisticRegression.
    """
    @property
    def coefficients(self):
        """
        Model coefficients of binomial logistic regression.
        An exception is thrown in the case of multinomial logistic regression.
        """
        return self._call_java("coefficients")

    @property
    def intercept(self):
        """
        Model intercept of binomial logistic regression.
        An exception is thrown in the case of multinomial logistic regression.
        """
        return self._call_java("intercept")

    @property
    def coefficientMatrix(self):
        """
        Model coefficients.
        """
        return self._call_java("coefficientMatrix")

    @property
    def interceptVector(self):
        """
        Model intercept.
        """
        return self._call_java("interceptVector")

    @property
    def summary(self):
        """
        Gets summary (e.g. accuracy/precision/recall, objective history, total iterations) of model
        trained on the training set. An exception is thrown if `trainingSummary is None`.
        """
        if self.hasSummary:
            java_lrt_summary = self._call_java("summary")
            if self.numClasses <= 2:
                return BinaryLogisticRegressionTrainingSummary(java_lrt_summary)
            else:
                return LogisticRegressionTrainingSummary(java_lrt_summary)
        else:
            raise RuntimeError("No training summary available for this %s" %
                               self.__class__.__name__)

    @property
    def hasSummary(self):
        """
        Indicates whether a training summary exists for this model
        instance.
        """
        return self._call_java("hasSummary")

    def evaluate(self, dataset):
        """
        Evaluates the model on a test dataset.

        :param dataset:
          Test dataset to evaluate model on, where dataset is an
          instance of :py:class:`pyspark.sql.DataFrame`
        """
        if not isinstance(dataset, DataFrame):
            raise ValueError("dataset must be a DataFrame but got %s." % type(dataset))
        java_blr_summary = self._call_java("evaluate", dataset)
        return BinaryLogisticRegressionSummary(java_blr_summary)


class LogisticRegressionSummary(JavaWrapper):
    """

    Abstraction for Logistic Regression Results for a given model.
    """

    @property
    def predictions(self):
        """
        Dataframe outputted by the model's `transform` method.
        """
        return self._call_java("predictions")

    @property
    def probabilityCol(self):
        """
        Field in "predictions" which gives the probability
        of each class as a vector.
        """
        return self._call_java("probabilityCol")

    @property
    def predictionCol(self):
        """
        Field in "predictions" which gives the prediction of each class.
        """
        return self._call_java("predictionCol")

    @property
    def labelCol(self):
        """
        Field in "predictions" which gives the true label of each
        instance.
        """
        return self._call_java("labelCol")

    @property
    def featuresCol(self):
        """
        Field in "predictions" which gives the features of each instance
        as a vector.
        """
        return self._call_java("featuresCol")

    @property
    def labels(self):
        """
        Returns the sequence of labels in ascending order. This order matches the order used
        in metrics which are specified as arrays over labels, e.g., truePositiveRateByLabel.

        Note: In most cases, it will be values {0.0, 1.0, ..., numClasses-1}, However, if the
        training set is missing a label, then all of the arrays over labels
        (e.g., from truePositiveRateByLabel) will be of length numClasses-1 instead of the
        expected numClasses.
        """
        return self._call_java("labels")

    @property
    def truePositiveRateByLabel(self):
        """
        Returns true positive rate for each label (category).
        """
        return self._call_java("truePositiveRateByLabel")

    @property
    def falsePositiveRateByLabel(self):
        """
        Returns false positive rate for each label (category).
        """
        return self._call_java("falsePositiveRateByLabel")

    @property
    def precisionByLabel(self):
        """
        Returns precision for each label (category).
        """
        return self._call_java("precisionByLabel")

    @property
    def recallByLabel(self):
        """
        Returns recall for each label (category).
        """
        return self._call_java("recallByLabel")

    def fMeasureByLabel(self, beta=1.0):
        """
        Returns f-measure for each label (category).
        """
        return self._call_java("fMeasureByLabel", beta)

    @property
    def accuracy(self):
        """
        Returns accuracy.
        (equals to the total number of correctly classified instances
        out of the total number of instances.)
        """
        return self._call_java("accuracy")

    @property
    def weightedTruePositiveRate(self):
        """
        Returns weighted true positive rate.
        (equals to precision, recall and f-measure)
        """
        return self._call_java("weightedTruePositiveRate")

    @property
    def weightedFalsePositiveRate(self):
        """
        Returns weighted false positive rate.
        """
        return self._call_java("weightedFalsePositiveRate")

    @property
    def weightedRecall(self):
        """
        Returns weighted averaged recall.
        (equals to precision, recall and f-measure)
        """
        return self._call_java("weightedRecall")

    @property
    def weightedPrecision(self):
        """
        Returns weighted averaged precision.
        """
        return self._call_java("weightedPrecision")

    def weightedFMeasure(self, beta=1.0):
        """
        Returns weighted averaged f-measure.
        """
        return self._call_java("weightedFMeasure", beta)


@inherit_doc
class LogisticRegressionTrainingSummary(LogisticRegressionSummary):
    """

    Abstraction for multinomial Logistic Regression Training results.
    Currently, the training summary ignores the training weights except
    for the objective trace.

    """

    @property
    def objectiveHistory(self):
        """
        Objective function (scaled loss + regularization) at each
        iteration.
        """
        return self._call_java("objectiveHistory")

    @property
    def totalIterations(self):
        """
        Number of training iterations until termination.
        """
        return self._call_java("totalIterations")


@inherit_doc
class BinaryLogisticRegressionSummary(LogisticRegressionSummary):
    """

    Binary Logistic regression results for a given model.

    """

    @property
    def roc(self):
        """
        Returns the receiver operating characteristic (ROC) curve,
        which is a Dataframe having two fields (FPR, TPR) with
        (0.0, 0.0) prepended and (1.0, 1.0) appended to it.

        .. seealso:: `Wikipedia reference \
        <http://en.wikipedia.org/wiki/Receiver_operating_characteristic>`_

        """
        return self._call_java("roc")

    @property
    def areaUnderROC(self):
        """
        Computes the area under the receiver operating characteristic
        (ROC) curve.

        """
        return self._call_java("areaUnderROC")

    @property
    def pr(self):
        """
        Returns the precision-recall curve, which is a Dataframe
        containing two fields recall, precision with (0.0, 1.0) prepended
        to it.

        """
        return self._call_java("pr")

    @property
    def fMeasureByThreshold(self):
        """
        Returns a dataframe with two fields (threshold, F-Measure) curve
        with beta = 1.0.

        """
        return self._call_java("fMeasureByThreshold")

    @property
    def precisionByThreshold(self):
        """
        Returns a dataframe with two fields (threshold, precision) curve.
        Every possible probability obtained in transforming the dataset
        are used as thresholds used in calculating the precision.

        """
        return self._call_java("precisionByThreshold")

    @property
    def recallByThreshold(self):
        """
        Returns a dataframe with two fields (threshold, recall) curve.
        Every possible probability obtained in transforming the dataset
        are used as thresholds used in calculating the recall.

        """
        return self._call_java("recallByThreshold")


@inherit_doc
class BinaryLogisticRegressionTrainingSummary(BinaryLogisticRegressionSummary,
                                              LogisticRegressionTrainingSummary):
    """

    Binary Logistic regression training results for a given model.

    """
    pass

@inherit_doc
class LinearSVC(JavaEstimator, HasFeaturesCol, HasLabelCol, HasPredictionCol, HasMaxIter,
                HasRegParam, HasTol, HasRawPredictionCol, HasWeightCol, JavaMLWritable, JavaMLReadable):
    """

    `Linear SVM Classifier <https://en.wikipedia.org/wiki/Support_vector_machine#Linear_SVM>`_


    >>> from pyspark.sql import Row
    >>> from pyspark.ml.linalg import Vectors
    >>> df = sc.parallelize([
    ...     Row(label=1.0, features=Vectors.dense(1.0, 1.0, 1.0)),
    ...     Row(label=0.0, features=Vectors.dense(1.0, 2.0, 3.0))]).toDF()
    >>> svm = LinearSVC(maxIter=5, regParam=0.01)
    >>> model = svm.fit(df)
    >>> model.coefficients
    DenseVector([0.0, -0.2792, -0.1833])
    >>> model.intercept
    1.0206118982229047
    >>> model.numClasses
    2
    >>> model.numFeatures
    3
    >>> test0 = sc.parallelize([Row(features=Vectors.dense(-1.0, -1.0, -1.0))]).toDF()
    >>> result = model.transform(test0).head()
    >>> result.prediction
    1.0
    >>> result.rawPrediction
    DenseVector([-1.4831, 1.4831])
    >>> svm_path = temp_path + "/svm"
    >>> svm.save(svm_path)
    >>> svm2 = LinearSVC.load(svm_path)
    >>> svm2.getMaxIter()
    5
    >>> model_path = temp_path + "/svm_model"
    >>> model.save(model_path)
    >>> model2 = LinearSVCModel.load(model_path)
    >>> model.coefficients[0] == model2.coefficients[0]
    True
    >>> model.intercept == model2.intercept
    True

    """

    verbose = Param(Params._dummy(), "verbose", "Flag for indicating if the training loss will be printed at each epoch.",
                    typeConverter=TypeConverters.toBoolean)

    def setVerbose(self, value):
        """
        Sets the value of :py:attr:`verbose`.
        """
        return self._set(verbose=value)

    def getVerbose(self):
        """
        Gets the value of verbose or its default value.
        """
        return self.getOrDefault(self.verbose)

    gpuMemLimit = Param(Params._dummy(), "gpuMemLimit", "Limit of the GPU memory. If set to the default value 0, the maximum possible memory is used.",
                        typeConverter=TypeConverters.toInt)

    def setGpuMemLimit(self, value):
        """
        Sets the value of :py:attr:`gpuMemLimit`.
        """
        return self._set(gpuMemLimit=value)

    def getGpuMemLimit(self):
        """
        Gets the value of gpuMemLimit or its default value.
        """
        return self.getOrDefault(self.gpuMemLimit)

    nthreads = Param(Params._dummy(), "nthreads", "Number of threads to be used. default : -1 meaning that n_threads=256 if GPU is enabled, else 1",
                     typeConverter=TypeConverters.toInt)

    def setNthreads(self, value):
        """
        Sets the value of :py:attr:`nthreads`.
        """
        return self._set(nthreads=value)

    def getNthreads(self):
        """
        Gets the value of nthreads or its default value.
        """
        return self.getOrDefault(self.nthreads)

    balanced = Param(Params._dummy(), "balanced", "If set to False, all classes will have weight 1.",
                     typeConverter=TypeConverters.toBoolean)

    def setBalanced(self, value):
        """
        Sets the value of :py:attr:`balanced`.
        """
        return self._set(balanced=value)

    def getBalanced(self):
        """
        Gets the value of balanced or its default value.
        """
        return self.getOrDefault(self.balanced)

    dual = Param(Params._dummy(), "dual", "Dual or Primal formulation. Recommendation: if n_samples > n_features use dual=True.",
                 typeConverter=TypeConverters.toBoolean)

    def setDual(self, value):
        """
        Sets the value of :py:attr:`dual`.
        """
        return self._set(dual=value)

    def getDual(self):
        """
        Gets the value of dual or its default value.
        """
        return self.getOrDefault(self.dual)

    useGpu = Param(Params._dummy(), "useGpu", "Flag for indicating the hardware platform used for training. If True, the training is performed using the GPU. If False, the training is performed using the CPU.",
                   typeConverter=TypeConverters.toBoolean)

    def setUseGpu(self, value):
        """
        Sets the value of :py:attr:`useGpu`.
        """
        return self._set(useGpu=value)

    def getUseGpu(self):
        """
        Gets the value of useGpu or its default value.
        """
        return self.getOrDefault(self.useGpu)

    trainingHistory = Param(Params._dummy(), "trainingHistory", "Flag for indicating the amount of information to collect and returned by the fit function. default: 0, no information is returned, 1, to obtain summary statistics at the end of training, 2, to obtain a complete set of statistics for the entire training procedure",
                            typeConverter=TypeConverters.toInt)

    def setTrainingHistory(self, value):
        """
        Sets the value of :py:attr:`trainingHistory`.
        """
        return self._set(trainingHistory=value)

    def getTrainingHistory(self):
        """
        Gets the value of trainingHistory or its default value.
        """
        return self.getOrDefault(self.trainingHistory)

    @keyword_only
    def __init__(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0,
                 maxIter=1000, regParam=1.0, tol=1e-3, rawPredictionCol="rawPrediction",
                 standardization=True, weightCol=None,
                 useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):
        """
        __init__(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0,\
                 maxIter=1000, regParam=1.0, tol=1e-3, rawPredictionCol="rawPrediction", \
                 standardization=True, weightCol=None, \
                 useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):
        """
        super(LinearSVC, self).__init__()
        self._java_obj = self._new_java_obj(
            "com.ibm.snap.ml.SnapLinearSVC", self.uid)
        self._setDefault(trainingHistory=0, maxIter=1000, regParam=1.0, tol=1e-3, useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(self, featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0,
                  maxIter=1000, regParam=1.0, tol=1e-3, rawPredictionCol="rawPrediction",
                  standardization=True, weightCol=None,
                  useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):
        """
        setParams(featuresCol="features", labelCol="label", predictionCol="prediction", trainingHistory=0, \
                 maxIter=1000, regParam=1.0, tol=1e-3, rawPredictionCol="rawPrediction", \
                 standardization=True, weightCol=None, \
                 useGpu=False, dual=True, balanced=False, nthreads=-1, gpuMemLimit=0, verbose=False):
        Sets params for Linear SVM Classifier.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def _create_model(self, java_model):
        return LinearSVCModel(java_model)


class LinearSVCModel(JavaModel, JavaClassificationModel, JavaMLWritable, JavaMLReadable):
    """

    Model fitted by LinearSVC.

    """

    @property
    def coefficients(self):
        """
        Model coefficients of Linear SVM Classifier.
        """
        return self._call_java("coefficients")

    @property
    def intercept(self):
        """
        Model intercept of Linear SVM Classifier.
        """
        return self._call_java("intercept")

    @property
    def summary(self):
        """
        Gets summary of model on
        training set. An exception is thrown if
        `trainingSummary is None`.
        """
        if self.hasSummary:
            java_lrt_summary = self._call_java("summary")
            return LinearSVCTrainingSummary(java_lrt_summary)
        else:
            raise RuntimeError("No training summary available for this %s" %
                               self.__class__.__name__)

    @property
    def hasSummary(self):
        """
        Indicates whether a training summary exists for this model
        instance.
        """
        return self._call_java("hasSummary")

class LinearSVCSummary(JavaWrapper):
    """

    Abstraction for Logistic Regression Results for a given model.
    """

    @property
    def predictions(self):
        """
        Dataframe outputted by the model's `transform` method.
        """
        return self._call_java("predictions")

    @property
    def predictionCol(self):
        """
        Field in "predictions" which gives the prediction of each class.
        """
        return self._call_java("predictionCol")

    @property
    def labelCol(self):
        """
        Field in "predictions" which gives the true label of each
        instance.
        """
        return self._call_java("labelCol")

    @property
    def featuresCol(self):
        """
        Field in "predictions" which gives the features of each instance
        as a vector.
        """
        return self._call_java("featuresCol")

    @property
    def labels(self):
        """
        Returns the sequence of labels in ascending order. This order matches the order used
        in metrics which are specified as arrays over labels, e.g., truePositiveRateByLabel.

        Note: In most cases, it will be values {0.0, 1.0, ..., numClasses-1}, However, if the
        training set is missing a label, then all of the arrays over labels
        (e.g., from truePositiveRateByLabel) will be of length numClasses-1 instead of the
        expected numClasses.
        """
        return self._call_java("labels")

    @property
    def truePositiveRateByLabel(self):
        """
        Returns true positive rate for each label (category).
        """
        return self._call_java("truePositiveRateByLabel")

    @property
    def falsePositiveRateByLabel(self):
        """
        Returns false positive rate for each label (category).
        """
        return self._call_java("falsePositiveRateByLabel")

    @property
    def precisionByLabel(self):
        """
        Returns precision for each label (category).
        """
        return self._call_java("precisionByLabel")

    @property
    def recallByLabel(self):
        """
        Returns recall for each label (category).
        """
        return self._call_java("recallByLabel")

    def fMeasureByLabel(self, beta=1.0):
        """
        Returns f-measure for each label (category).
        """
        return self._call_java("fMeasureByLabel", beta)

    @property
    def accuracy(self):
        """
        Returns accuracy.
        (equals to the total number of correctly classified instances
        out of the total number of instances.)
        """
        return self._call_java("accuracy")

    @property
    def weightedTruePositiveRate(self):
        """
        Returns weighted true positive rate.
        (equals to precision, recall and f-measure)
        """
        return self._call_java("weightedTruePositiveRate")

    @property
    def weightedFalsePositiveRate(self):
        """
        Returns weighted false positive rate.
        """
        return self._call_java("weightedFalsePositiveRate")

    @property
    def weightedRecall(self):
        """
        Returns weighted averaged recall.
        (equals to precision, recall and f-measure)
        """
        return self._call_java("weightedRecall")

    @property
    def weightedPrecision(self):
        """
        Returns weighted averaged precision.
        """
        return self._call_java("weightedPrecision")

    def weightedFMeasure(self, beta=1.0):
        """
        Returns weighted averaged f-measure.
        """
        return self._call_java("weightedFMeasure", beta)


@inherit_doc
class LinearSVCTrainingSummary(LinearSVCSummary):
    """

    Linear regression training results. Currently, the training summary ignores the
    training weights except for the objective trace.

    """

    @property
    def objectiveHistory(self):
        """
        Objective function (scaled loss + regularization) at each
        iteration.

        .. seealso:: :py:attr:`LinearRegression.solver`
        """
        return self._call_java("objectiveHistory")

    @property
    def totalIterations(self):
        """
        Number of training iterations until termination.

        .. seealso:: :py:attr:`LinearRegression.solver`
        """
        return self._call_java("totalIterations")
