# criteo-kaggle example

In this example we will train a Logistic Regression on the criteo-kaggle dataset, using both sparkML and snap-ml-spark. 

The criteo-kaggle dataset is from the [Display Advertising Challenge](https://www.kaggle.com/c/criteo-display-ad-challenge). 

The contest is described as follows:

> Display advertising is a billion dollar effort and one of the central uses of machine learning on the Internet. However, its data and methods are usually kept under lock and key. In this research competition, CriteoLabs is sharing a week’s worth of data for you to develop models predicting ad click-through rate (CTR). Given a user and the page he is visiting, what is the probability that he will click on a given ad?


## Download the Data

Download and extract the data from the LIBSVM repository:
```bash
mkdir data
cd data
wget https://s3-us-west-2.amazonaws.com/criteo-public-svm-data/criteo.kaggle2014.svm.tar.gz
tar xzf criteo.kaggle2014.svm.tar.gz
cd ../
```

After the file extraction one should see the two files:
(1) training set: `criteo.kaggle2014.train.svm`
(2) test set: `criteo.kaggle2014.test.svm`

The file (2) can be used as a test set only for the kaggle submission and thus in the next step we will keep only the train set in order to split it again in train/test sets in the next section.
```bash
cd data
rm criteo.kaggle2014.test.svm
mv criteo.kaggle2014.train.svm criteo.kaggle2014
cd ../
```


## Data Pre-processing

In this pre-processing step we will load the data and make a train/test split. Following, we will save the files in libsvm format so that we can also read them from spark ML and perform a performance comparison against `snap_ml_spark`:
Firstly, we load the data from `libsvm` format:
```python
from sklearn.datasets import load_svmlight_file
X,y = load_svmlight_file("data/criteo.kaggle2014.svm")
```

Next, we make a train/test split of the data:
```python
from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25, random_state=42)
```
Then we re-write the data to libsvm format.
```python
from sklearn.datasets import dump_svmlight_file
print('Writing training data to libsvm format')
dump_svmlight_file(X_train,y_train, filename+'-train.libsvm', zero_based=False)
print('Writing test data to libsvm format')
dump_svmlight_file(X_test,y_test, filename+'-test.libsvm', zero_based=False)
```

The entire pre-processing code can be run by executing: `python preprocess-criteo45m.py`

## Training and Evaluating a Logistic Regression Model

After running the data pre-processing steps above, we are ready to perform our benchmarking:

Firstly we need to import the pySpark modules, set the configuration and disable manually the LOG INFO printed by spark

```python
from pyspark import SparkConf, SparkContext
from pyspark.sql.session import SparkSession

conf = SparkConf()\
.set("spark.solver.count",args.num_partitions)\
.set("spark.cores.max",args.num_partitions)\
.setAppName("Spark-criteoKaggle2014-SparkComparison")\
.set("spark.executor.extraJavaOptions","-XX:MaxDirectMemorySize=8192m")\
.set("spark.driver.maxResultSize",8192)
# Create spark config
sc = SparkContext(conf=conf)
# Configure Spark
spark = SparkSession(sc)

log4j = sc._jvm.org.apache.log4j
log4j.LogManager.getRootLogger().setLevel(log4j.Level.ERROR)
```
The configuration can vary depending on your Spark cluster

Then, our next step is to define the data paths.
Please note that if you wish to run Spark on a real distributed environment, then you must set the absolute data paths of the train/test sets and
also have access to the data from all the connected machines that will be used (by replicating the data or by providing a distributed file system).
```python
filename = 'data/criteo.kaggle2014.'
train_filename = filename + 'train.svm'
test_filename = filename + 'test.svm'
```

Next, starting from the `snap_ml_spark` library first, our first step is to read the libsvm train and test data:
```python
from snap_ml_spark import DatasetReader

from snap_ml_spark import LogisticRegression as snapml_LogisticRegression
from snap_ml_spark.Metrics import accuracy, logisticLoss

n_features_ = 1000000

train_data = DatasetReader().setFormat("libsvm").setNumFt(n_features_).load(train_filename)

test_data = DatasetReader().setFormat("libsvm").setNumFt(n_features_).load(test_filename)
```
Note that since we are reading the data from the libsvm format we need to specify the number of features of the datasets

Then, we initialize our `snap_ml_spark` LogisticRegression constructor and we set the number of coordinate descent epochs to 6 and the snapml dual LR regularizer to 10.0.
We leave the rest to the default values (GPUs are disabled). After the initialization, all we need to do is to call the `fit()` function in order to train our dual Logistic Regression model (we also time it):
```python
snapml_regularizer = 10.0
snapml_lr = snapml_LogisticRegression(max_iter=6, regularizer=snapml_regularizer, verbose=False, dual=True, use_gpu=False, n_threads=-1) # 10

snapml_t0 = time.time()
snapml_lr.fit(train_data)
snapml_time_cpu = time.time() - snapml_t0
```

As a last step, we make our predictions on the test data, we compute the accuracy metric and we print off the results (with the time needed to achieve them):
```python
pred = snapml_lr.predict(test_data)

snapml_accuracy_cpu = accuracy(pred)

print('snapML CPU accuracy: %.4f' %snapml_accuracy_cpu, ", CPU time: %.2f" % snapml_time_cpu, 's')
```

Then, we perform the same steps for the GPU snap_ml_spark Logistic Regression algorithm. Note that the `use_gpu` parameter in the Logistic Regression constructor is set to `False`, so we need to set it to `True`.
```python
snapml_lr = snapml_LogisticRegression(max_iter=45, regularizer=snapml_regularizer, verbose=False, dual=True, use_gpu=True, n_threads=-1) # 40

snapml_t0 = time.time()
snapml_lr.fit(train_data)
snapml_time_gpu = time.time() - snapml_t0

pred = snapml_lr.predict(test_data)

snapml_accuracy_gpu = accuracy(pred)

print('snapML GPU accuracy: %.4f' %snapml_accuracy_gpu, ", GPU time: %.2f" % snapml_time_gpu, 's')
```

Finally, we perform the same steps using the Spark ML library and the same pySpark configuration
```python
from pyspark.ml.classification import LogisticRegression as sparkml_LogisticRegression

train_data = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(train_filename)
test_data = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(test_filename)
n_examples = train_data.count()
	
sparkml_lr = sparkml_LogisticRegression(fitIntercept=False, regParam=snapml_regularizer/n_examples, standardization=False)

sparkml_t0 = time.time()
sparkml_lr_model = sparkml_lr.fit(train_data)
sparkml_time = time.time() - sparkml_t0

predictions = sparkml_lr_model.transform(test_data)

from pyspark.ml.evaluation import MulticlassClassificationEvaluator
evaluator = MulticlassClassificationEvaluator(labelCol="label", predictionCol="prediction", metricName="accuracy")
sparkml_accuracy = evaluator.evaluate(predictions)

print('Spark ML', evaluator.getMetricName(),'=', sparkml_accuracy,", time: %.2f" % sparkml_time, 's')
```
Note that we set the `fitIntercept` and the `standardization` parameters to `False` and the regularizer to `snapml_regularizer/n_examples` in order to make the spark training equivalent to the snapML training.

After the final training and prediction on the train/test data using the Spark ML lib, we finally print off all the results:
```python
# Print off all Results
print('\nsnapML  CPU accuracy: %.4f' %snapml_accuracy_cpu, ", CPU time: %.2f" % snapml_time_cpu, 's')
print('snapML  GPU accuracy: %.4f' %snapml_accuracy_gpu, ", GPU time: %.2f" % snapml_time_gpu, 's')
print('sparkML lib accuracy: %.4f' % sparkml_accuracy, ",    time: %.2f" % sparkml_time, 's')
```

One can run the entire example by executing:
```bash
spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-criteo45m/example-criteo45m.py --data_path path/to/data

Note: By default the location of --data_path is current directory. Parent directory of the directory 'data/' is to be given as parameter to --data_path option.

Command to run example using GPUs:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-criteo45m/example-criteo45m.py --use_gpu

Command to run example with specified number of partitions:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-criteo45m/example-criteo45m.py --num_partitions 8

Note: Default value of num_partitions is 4. num_partitions can be 8, if user wants to use 8 GPUs.
```
One should see the following output on a **P9+V100 machine with 4 gpus**:
```
[...]
snapML  GPU accuracy: 0.7856 , GPU time: 25.26 s
sparkML     accuracy: 0.7850 ,    time: 2442.81 s
```
Please keep in mind that the `snap_ml_spark` CPU algorithm converges significantly faster than the snap_ml_spark GPU algorithm with respect to the number of epochs.
The reported times can vary depending on system architecture.

## Duality

In this example we additionally provide one example when we train our Logistic Regression model on the same dataset but using the primal and the dual formulation.
The steps are identical with the aforementioned steps.
One should run the following python script 
```bash
python preprocess-criteo45m-duality.py
```
to preprocess the data and then the following command to execute the duality example: 
```bash
spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-criteo45m/example-criteo45m-duality.py  --data_path path/to/data

Note: By default the location of --data_path is current directory. Parent directory of the directory 'data/' is to be given as parameter to --data_path option.

Command to run example using GPUs:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-criteo45m/example-criteo45m-duality.py --use_gpu

Command to run example with specified number of partitions:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-criteo45m/example-criteo45m-duality.py --num_partitions 8

Note: Default value of num_partitions is 4. num_partitions can be 8, if user wants to use 8 GPUs.
```
&copy; Copyright IBM Corporation 2018, 2019
